package org.bouncycastle.asn1;

import java.math.BigInteger;
import java.io.*;

public class DERInputStream
    extends FilterInputStream implements DERTags
{
    public DERInputStream(
        InputStream is)
    {
        super(is);
    }

    private int readLength()
        throws IOException
    {
        int length = read();
        if (length < 0)
        {
            throw new IOException("EOF found when length expected");
        }

        if (length > 127)
        {
            int size = length & 0x7f;

            length = 0;
            for (int i = 0; i < size; i++)
            {
                int next = read();

                if (next < 0)
                {
                    throw new IOException("EOF found reading length");
                }

                length = (length << 8) + next;
            }
        }

        return length;
    }

    private void readFully(
        byte[]  bytes)
        throws IOException
    {
        int     left = bytes.length;

        if (left == 0)
        {
            return;
        }

        while ((left -= read(bytes, bytes.length - left, left)) != 0)
        {
            ;
        }
    }

    public DERObject readObject()
        throws IOException
    {
        int tag = read();
        if (tag == -1)
        {
            throw new EOFException();
        }

        int     length = readLength();
        byte[]  bytes = new byte[length];

        readFully(bytes);

        switch (tag)
        {
        case NULL:
            return null;   
        case SEQUENCE | CONSTRUCTED:
            ByteArrayInputStream    bIn = new ByteArrayInputStream(bytes);
            DERInputStream          dIn = new DERInputStream(bIn);
            DERConstructedSequence  seq = new DERConstructedSequence();

            try
            {
                for (;;)
                {
                    DERObject   obj = dIn.readObject();

                    seq.addObject(obj);
                }
            }
            catch (EOFException ex)
            {
                return seq;
            }
        case SET | CONSTRUCTED:
            bIn = new ByteArrayInputStream(bytes);
            dIn = new DERInputStream(bIn);

            return new DERSet(dIn.readObject());
        case BOOLEAN:
            return new DERBoolean(bytes);
        case INTEGER:
            return new DERInteger(bytes);
        case OBJECT_IDENTIFIER:
            int             head = bytes[0] & 0xff;
            StringBuffer    objId = new StringBuffer();
    
            objId.append(Integer.toString(head / 40));
            objId.append('.');
            objId.append(Integer.toString(head % 40));
            
            int value = 0;
    
            for (int i = 1; i != length; i++)
            {
                int b = bytes[i] & 0xff;
    
                value = value * 128 + (b & 0x7f);
                if ((b & 128) == 0)             // end of number reached
                {
                    objId.append('.');
                    objId.append(Integer.toString(value));
                    value = 0;
                }
            }
    
            return new DERObjectIdentifier(objId.toString());
        case BIT_STRING:
            int     padBits = bytes[0];
            byte[]  data = new byte[length - 1];

            System.arraycopy(bytes, 1, data, 0, length - 1);

            return new DERBitString(data, padBits);
        case PRINTABLE_STRING:
            return new DERPrintableString(bytes);
        case IA5_STRING:
            return new DERIA5String(bytes);
        case OCTET_STRING:
            return new DEROctetString(bytes);
        case UTC_TIME:
            return new DERUTCTime(new String(bytes));
        default:
            if ((tag & (TAGGED | CONSTRUCTED)) != 0)  
            {
                if (length == 0)        // empty tag!
                {
                    return new DERTaggedObject(tag & 0x0f);
                }

                bIn = new ByteArrayInputStream(bytes);
                dIn = new DERInputStream(bIn);

                            // with tagged object tag number is bottom 4 bits
                return new DERTaggedObject(tag & 0x0f, dIn.readObject());
            }

            return new DERUnknownTag(tag, bytes);
        }
    }
}
